from __future__ import annotations

from collections.abc import Callable, Iterator
from functools import cmp_to_key
from math import hypot, isnan, pi, sin

from ..array import argpass
from ..types import T
from .context import Context
from .delaunator import Delaunator
from .path import Path
from .polygon import Polygon
from .voronoi import Voronoi

TAU = 2 * pi


def point_x(p: tuple[float, float]) -> float:
    return p[0]


def point_y(p: tuple[float, float]) -> float:
    return p[1]


def collinear(d: Delaunator) -> bool:
    triangles = d.triangles
    coords = d.coords
    for i in range(0, len(triangles), 3):
        a = 2 * triangles[i]
        b = 2 * triangles[i + 1]
        c = 2 * triangles[i + 2]
        cross = (coords[c] - coords[a]) * (coords[b + 1] - coords[a + 1]) - (
            coords[b] - coords[a]
        ) * (coords[c + 1] - coords[a + 1])
        if cross > 1e-10:
            return False
    return True


class Delaunay:
    """
    Delaunay triangulation for the given flat array :math:`[x_0, y_0, x_1, y_1,
    \\ldots]` of points.

    Parameters
    ----------
    points : list[float]
        The coordinates of the points as an array :math:`[x_0, y_0, x_1, y_1,
        \\ldots]`.

    Attributes
    ----------
    inedges : list[int]
        The incoming halfedge indexes :math:`[e_0, e_1, e_2, \\ldots]`. For
        each point :code:`i`, :code:`inedges[i]` is the halfedge index
        :code:`e` of an incoming halfedge. For coincident points, the halfedge
        index is :code:`-1`; for points on the convex hull, the incoming
        halfedge is on the convex hull; for other points, the choice of
        incoming halfedge is arbitrary. The inedges table can be used to
        traverse the Delaunay triangulation.
    points : list[float]
        The coordinates of the points as an array :math:`[x_0, y_0, x_1, y_1,
        \\ldots]`.
    halfedges : list[int]
        The halfedge indexes as an Int32Array :math:`[j_0, j_1, \\ldots]`. For
        each index :math:`0 \\le i \\lt \\text{halfedges.length}`, there is a
        halfedge from triangle vertex :code:`j = halfedges[i]` to triangle
        vertex :code:`i`. Equivalently, this means that triangle
        :math:`\\lfloor i / 3 \\rfloor` is adjacent to triangle :math:`\\lfloor
        j / 3 \\rfloor`. If j is negative, then triangle :math:`\\lfloor i / 3
        \\rfloor` is an exterior triangle on the convex hull.
    hull : list[int]
        A list of point indexes that form the convex hull in counterclockwise
        order. If the points are collinear, returns them ordered.
    triangles : list[int]
        The triangle vertex indexes :math:`[i_0, j_0, k_0, i_1, j_1, k_1,
        \\ldots]`. Each contiguous triplet of indexes :code:`i`, :code:`j`,
        :code:`k` forms a counterclockwise triangle. The coordinates of the
        triangle's points can be found by going through
        :code:`delaunay.points`.
    """

    def __init__(self, points: list[float]):
        self._delaunator = Delaunator(points)
        self._hull_index = [0] * (len(points) // 2)

        self.inedges = [0] * (len(points) // 2)
        self.points = self._delaunator.coords

        self.collinear = None
        self.halfedges = None
        self.hull = None
        self.triangles = None

        self._initialize()

    @classmethod
    def from_points(
        cls,
        points: list[T],
        fx: Callable[[T, int, list[T]], float] = point_x,
        fy: Callable[[T, int, list[T]], float] = point_y,
    ) -> Delaunay:
        """
        Returns the Delaunay triangulation for the given array or iterable of
        points. If :code:`fx` and :code:`fy` are not specified, then points is
        assumed to be an array of two-element arrays of numbers: :math:`[[x_0,
        y_0], [x_1, y_1], \\ldots]`.

        Parameters
        ----------
        points : list[T]
            List of points
        fx : Callable[[T, int, list[T]], float]
            X-coordinate accessor
        fy : Callable[[T, int, list[T]], float]
            Y-coordinate accessor

        Returns
        -------
        Delaunay
            Delaunay object
        """
        n = len(points)
        array = [None] * (n * 2)
        fx = argpass(fx)
        fy = argpass(fy)
        for i in range(n):
            p = points[i]
            array[i * 2] = fx(p, i, points)
            array[i * 2 + 1] = fy(p, i, points)
        return cls(array)

    def update(self) -> Delaunay:
        """
        Recomputes the triangulation after the points have been modified
        in-place.

        Returns
        -------
        Delaunay
            Itself
        """
        self._delaunator.update()
        self._initialize()
        return self

    def _initialize(self):
        d = self._delaunator
        points = self.points

        if d.hull and len(d.hull) > 2 and collinear(d):

            def compare(i, j):
                return (
                    points[2 * i] - points[2 * j]
                    or points[2 * i + 1] - points[2 * j + 1]
                )

            self.collinear = sorted(range(len(points) // 2), key=cmp_to_key(compare))
            e = self.collinear[0] << 1
            f = self.collinear[-1] << 1
            r = 1e-9 * hypot(points[f] - points[e], points[f + 1] - points[e + 1])
            for i in range(len(points)):
                points[i] += r * sin(i + 0.5)
            self._delaunator = Delaunator(points)
        else:
            self.collinear = None

        halfedges = self.halfedges = self._delaunator.halfedges
        hull = self.hull = self._delaunator.hull
        triangles = self.triangles = self._delaunator.triangles
        inedges = self.inedges = [-1 for _ in self.inedges]
        hull_index = self._hull_index = [-1 for _ in self._hull_index]

        for e in range(len(halfedges)):
            p = triangles[e - 2 if e % 3 == 2 else e + 1]
            if halfedges[e] == -1 or inedges[p] == -1:
                inedges[p] = e

        for i, h in enumerate(hull):
            hull_index[h] = i

        if 0 < len(hull) <= 2:
            self.triangles = [-1] * 3
            self.halfedges = [-1] * 3
            self.triangles[0] = hull[0]
            inedges[hull[0]] = 1
            if len(hull) == 2:
                inedges[hull[1]] = 0
                self.triangles[1] = hull[1]
                self.triangles[2] = hull[1]

    def voronoi(
        self, bounds: tuple[float, float, float, float] | None = None
    ) -> Voronoi:
        """
        Returns the Voronoi diagram for the given Delaunay triangulation. When
        rendering, the diagram will be clipped to the specified :code:`bounds =
        [xmin, ymin, xmax, ymax]`.

        Parameters
        ----------
        bounds : tuple[float, float, float, float] | None
            :code:`[xmin, ymin, xmax, ymax]` values

        Returns
        -------
        Voronoi
            Voronoi object
        """
        return Voronoi(self, bounds)

    def neighbors(self, i: int) -> Iterator[int]:
        """
        Returns an iterable over the indexes of the neighboring points to the
        specified point :code:`i`. The iterable is empty if :code:`i` is a
        coincident point.

        Parameters
        ----------
        i : int
            Point index

        Returns
        -------
        Iterator[int]
            Indexes of the neighboring points
        """
        inedges = self.inedges
        hull = self.hull
        halfedges = self.halfedges
        _hull_index = self._hull_index
        triangles = self.triangles
        collinear = self.collinear

        if collinear:
            if i not in collinear:
                return
            k = self.collinear.index(i)
            if k > 0:
                yield self.collinear[k - 1]
            if k < len(self.collinear) - 1:
                yield self.collinear[k + 1]
            return

        e0 = inedges[i]
        if e0 == -1:
            return
        e = e0
        p0 = -1
        while True:
            p0 = triangles[e]
            yield p0
            e = e - 2 if e % 3 == 2 else e + 1
            if triangles[e] != i:
                return
            e = halfedges[e]
            if e == -1:
                p = hull[(_hull_index[i] + 1) % len(hull)]
                if p != p0:
                    yield p
                return
            if e == e0:
                break

    def find(self, x: float, y: float, i: int = 0) -> int:
        """
        Returns the index of the input point that is closest to the specified
        point :math:`(x, y)`. The search is started at the specified point
        :code:`i`. If :code:`i` is not specified, it defaults to zero.

        Parameters
        ----------
        x : float
            X-coordinate point
        y : float
            Y-coordinate point
        i : int
            Starting point index

        Returns
        -------
        int
            Index of the input point
        """
        if isnan(x) or isnan(y):
            return -1

        i0 = i
        c = self._step(i, x, y)
        while c >= 0 and c != i and c != i0:
            i = c
            c = self._step(i, x, y)
        return c

    def _step(self, i: int, x: float, y: float) -> int:
        inedges = self.inedges
        hull = self.hull
        _hull_index = self._hull_index
        halfedges = self.halfedges
        triangles = self.triangles
        points = self.points

        if len(points) < 2:
            return -1
        if inedges[i] == -1:
            return (i + 1) % (len(points) >> 1)

        c = i
        dc = pow(x - points[i * 2], 2) + pow(y - points[i * 2 + 1], 2)
        e0 = inedges[i]
        e = e0
        while True:
            t = triangles[e]
            dt = pow(x - points[t * 2], 2) + pow(y - points[t * 2 + 1], 2)
            if dt < dc:
                dc = dt
                c = t
            e = e - 2 if e % 3 == 2 else e + 1
            if triangles[e] != i:
                break
            e = halfedges[e]
            if e == -1:
                e = hull[(_hull_index[i] + 1) % len(hull)]
                if e != t:
                    if (pow(x - points[e * 2], 2) + pow(y - points[e * 2 + 1], 2)) < dc:
                        return e
                break
            if e == e0:
                break
        return c

    def render(self, context: Context | None = None) -> str | None:
        """
        Renders the edges of the Delaunay triangulation to the specified
        context. If a :code:`context` is not specified, an SVG path string is
        returned instead.

        Parameters
        ----------
        context : Context | None
            Context object

        Returns
        -------
        str | None
            SVG path string
        """
        if context is None:
            buffer = context = Path()
        else:
            buffer = None

        points = self.points
        halfedges = self.halfedges
        triangles = self.triangles

        for i, j in enumerate(halfedges):
            if j < i:
                continue
            ti = triangles[i] * 2
            tj = triangles[j] * 2
            context.move_to(points[ti], points[ti + 1])
            context.line_to(points[tj], points[tj + 1])

        self.render_hull(context)
        return None if buffer is None else str(buffer)

    def render_points(
        self,
        context: Context | None = None,
        r: float | None = None,
    ) -> str | None:
        """
        Renders the input points of the Delaunay triangulation to the specified
        context as circles with the specified radius. If :code:`radius` is not
        specified, it defaults to 2. If a :code:`context` is not specified, an
        SVG path string is returned instead.

        Parameters
        ----------
        context : Context | None
            Context object
        r : float | None
            Radius value

        Returns
        -------
        str | None
            SVG path string
        """
        if r is None and (
            context is None
            or not (
                hasattr(context, "move_to") and callable(getattr(context, "move_to"))
            )
        ):
            r = context
            context = None

        r = 2 if r is None else r
        if context is None:
            buffer = context = Path()
        else:
            buffer = None

        points = self.points

        n = len(points) & ~1
        for i in range(0, n, 2):
            x = points[i]
            y = points[i + 1]
            context.move_to(x + r, y)
            context.arc(x, y, r, 0, TAU)

        return None if buffer is None else str(buffer)

    def render_hull(self, context: Context | None = None) -> str | None:
        """
        Renders the convex hull of the Delaunay triangulation to the specified
        context. If a :code:`context` is not specified, an SVG path string is
        returned instead.

        Parameters
        ----------
        context : Context | None
            Context object

        Returns
        -------
        str | None
            SVG path string
        """
        if context is None:
            buffer = context = Path()
        else:
            buffer = None

        hull = self.hull
        points = self.points

        h = hull[0] * 2
        context.move_to(points[h], points[h + 1])
        for i in range(1, len(hull)):
            h = 2 * hull[i]
            context.line_to(points[h], points[h + 1])
        context.close_path()
        return None if buffer is None else str(buffer)

    def hull_polygon(self) -> list[tuple[float, float]] | None:
        """
        Returns the closed polygon :math:`[[x_0, y_0], [x_1, y_1], \\ldots,
        [x_0, y_0]]` representing the convex hull.

        Returns
        -------
        list[tuple[float, float]] | None
            Closed polygon
        """
        polygon = Polygon()
        self.render_hull(polygon)
        return polygon.value()

    def render_triangle(self, i: int, context: Context | None = None) -> str | None:
        """
        Renders triangle :code:`i` of the Delaunay triangulation to the
        specified context. If a :code:`context` is not specified, an SVG path
        string is returned instead.

        Parameters
        ----------
        i : int
            Triangle index
        context : Context | None
            Context object

        Returns
        -------
        str | None
            SVG path string
        """
        if context is None:
            buffer = context = Path()
        else:
            buffer = None

        points = self.points
        triangles = self.triangles

        i *= 3
        t0 = triangles[i] * 3
        t1 = triangles[i + 1] * 3
        t2 = triangles[i + 2] * 3
        context.move_to(points[t0], points[t0 + 1])
        context.line_to(points[t1], points[t1 + 1])
        context.line_to(points[t2], points[t2 + 1])
        context.close_path()

        return None if buffer is None else str(buffer)

    def triangle_polygons(self) -> Iterator[list[tuple[float, float]] | None]:
        """
        Returns an iterable over the polygons for each triangle, in order.

        Returns
        -------
        Iterator[list[tuple[float, float]] | None]
            Iterator over the polygons for each triangle
        """
        triangles = self.triangles
        for i in range(len(triangles) // 3):
            yield self.triangle_polygon(i)

    def triangle_polygon(self, i: int) -> list[tuple[float, float]] | None:
        """
        Returns the closed polygon :math:`[[x_0, y_0], [x_1, y_1], [x_2, y_2],
        [x_0, y_0]]` representing the triangle :code:`i`.


        Parameters
        ----------
        i : int
            Triangle index

        Returns
        -------
        list[tuple[float, float]] | None
            Closed polygon
        """
        polygon = Polygon()
        self.render_triangle(i, polygon)
        return polygon.value()



from __future__ import annotations

from collections.abc import Iterator
from math import floor, inf, isnan, nan
from typing import TypeVar

from .context import Context
from .path import Path
from .polygon import Polygon

Delaunay = TypeVar("Delaunay", bound="Delaunay")


def sign(x: float) -> float:
    if x == 0:
        return 0
    elif x < 0:
        return -1
    else:
        return 1


class Voronoi:
    """
    Voronoi diagram for the given Delaunay triangulation. When rendering, the
    diagram will be clipped to the specified :code:`bounds = [xmin, ymin, xmax,
    ymax]`.

    Parameters
    ----------
    delaunay : Delaunay
        Delaunay object
    bounds : tuple[float, float, float, float] | None
        The bounds of the viewport :code:`[xmin, ymin, xmax, ymax]` for
        rendering the Voronoi diagram. If :code:`bounds` is not specified, it
        defaults to :code:`[0, 0, 960, 500]`

    Attributes
    ----------
    delaunay : Delaunay
        Delaunay object
    vectors : list[float]
        List :math:`[v_{x_0}, v_{y_0}, w_{x_0}, w_{y_0}, \\ldots]`  where each
        non-zero quadruple describes an open (infinite) cell on the outer hull,
        giving the directions of two open half-lines.
    circumcenters : list[float]
        The circumcenters of the Delaunay triangles :math:`[c_{x_0}, c_{y_0},
        c_{x_1}, c_{y_1}, \\ldots]`. Each contiguous pair of coordinates
        :math:`c_x`, :math:`c_y` is the circumcenter for the corresponding
        triangle. These circumcenters form the coordinates of the Voronoi cell
        polygons.
    xmin : float
        Bound of viewport
    xmax : float
        Bound of viewport
    ymin : float
        Bound of viewport
    ymax : float
        Bound of viewport
    """

    def __init__(
        self,
        delaunay: Delaunay,
        bounds: tuple[float, float, float, float] | None = None,
    ):
        if bounds is None:
            bounds = [0, 0, 960, 500]

        xmin, ymin, xmax, ymax = bounds
        if xmax < xmin or ymax < ymin:
            raise ValueError("Invalid bounds")

        self._circumcenters = [0.0] * (len(delaunay.points) * 2)

        self.delaunay = delaunay
        self.vectors = [0.0] * (len(delaunay.points) * 2)
        self.circumcenters = self._circumcenters
        self.xmax = xmax
        self.xmin = xmin
        self.ymax = ymax
        self.ymin = ymin

        self._initialize()

    def update(self) -> Voronoi:
        """
        Updates the Voronoi diagram and underlying triangulation after the
        points have been modified in-place — useful for Lloyd's relaxation.

        Returns
        -------
        Voronoi
            Itself
        """
        self.delaunay.update()
        self._initialize()
        return self

    def _initialize(self):
        points = self.delaunay.points
        hull = self.delaunay.hull
        triangles = self.delaunay.triangles

        circumcenters = self.circumcenters = self._circumcenters[
            0 : len(triangles) // 3 * 2
        ]
        bx = None
        by = None
        i = 0
        j = 0
        n = len(triangles)
        while i < n:
            t1 = triangles[i] * 2
            t2 = triangles[i + 1] * 2
            t3 = triangles[i + 2] * 2
            x1 = points[t1]
            y1 = points[t1 + 1]
            x2 = points[t2]
            y2 = points[t2 + 1]
            x3 = points[t3]
            y3 = points[t3 + 1]

            dx = x2 - x1
            dy = y2 - y1
            ex = x3 - x1
            ey = y3 - y1
            ab = (dx * ey - dy * ex) * 2

            if abs(ab) < 1e-9:
                if bx is None:
                    bx = by = 0
                    for k in hull:
                        bx += points[k * 2]
                        by += points[k * 2 + 1]
                    length = len(hull)
                    bx /= length
                    by /= length
                a = 1e9 * sign((bx - x1) * ey - (by - y1) * ex)
                x = (x1 + x3) / 2 - a * ey
                y = (y1 + y3) / 2 + a * ex
            else:
                d = 1 / ab
                bl = dx * dx + dy * dy
                cl = ex * ex + ey * ey
                x = x1 + (ey * bl - dy * cl) * d
                y = y1 + (dx * cl - ex * bl) * d
            circumcenters[j] = x
            circumcenters[j + 1] = y

            i += 3
            j += 2

        h = hull[-1]
        p0 = None
        p1 = h * 4
        x0 = None
        x1 = points[2 * h]
        y0 = None
        y1 = points[2 * h + 1]
        vectors = self.vectors = [0 for _ in self.vectors]
        for h in hull:
            p0 = p1
            x0 = x1
            y0 = y1
            p1 = h * 4
            x1 = points[2 * h]
            y1 = points[2 * h + 1]
            vectors[p0 + 2] = vectors[p1] = y0 - y1
            vectors[p0 + 3] = vectors[p1 + 1] = x1 - x0

    def render(self, context: Context | None = None) -> str | None:
        """
        Renders the mesh of Voronoi cells to the specified context. If a
        :code:`context` is not specified, an SVG path string is returned
        instead.

        Parameters
        ----------
        context : Context | None
            Context object

        Returns
        -------
        str | None
            SVG path string
        """
        if context is None:
            buffer = context = Path()
        else:
            buffer = None

        halfedges = self.delaunay.halfedges
        inedges = self.delaunay.inedges
        hull = self.delaunay.hull
        circumcenters = self.circumcenters
        vectors = self.vectors

        if len(hull) <= 1:
            return None

        for i, j in enumerate(halfedges):
            if j < i:
                continue
            ti = floor(i / 3) * 2
            tj = floor(j / 3) * 2
            xi = circumcenters[ti]
            yi = circumcenters[ti + 1]
            xj = circumcenters[tj]
            yj = circumcenters[tj + 1]
            self._render_segment(xi, yi, xj, yj, context)

        h0 = None
        h1 = hull[-1]
        for i in range(len(hull)):
            h0 = h1
            h1 = hull[i]
            t = floor(inedges[h1] / 3) * 2
            x = circumcenters[t]
            y = circumcenters[t + 1]
            v = h0 * 4
            p = self._project(x, y, vectors[v + 2], vectors[v + 3])
            if p:
                self._render_segment(x, y, p[0], p[1], context)
        return None if buffer is None else str(buffer)

    def render_bounds(self, context: Context | None = None) -> str | None:
        """
        Renders the viewport extent to the specified :code:`context`.
        Equivalent to :code:`context.rect(voronoi.xmin, voronoi.ymin,
        voronoi.xmax - voronoi.xmin, voronoi.ymax - voronoi.ymin)`. If a
        :code:`context` is not specified, an SVG path string is returned
        instead.

        Parameters
        ----------
        context : Context | None
            Context object

        Returns
        -------
        str | None
            SVG path string
        """
        if context is None:
            buffer = context = Path()
        else:
            buffer = None

        context.rect(
            self.xmin,
            self.ymin,
            self.xmax - self.xmin,
            self.ymax - self.ymin,
        )
        return None if buffer is None else str(buffer)

    def render_cell(self, i: int, context: Context | None = None) -> str | None:
        """
        Renders the cell with the specified index :code:`i` to the specified
        context. If a :code:`context` is not specified, an SVG path string is
        returned instead.

        Parameters
        ----------
        context : Context | None
            Context object
        i : int
           Cell index

        Returns
        -------
        str | None
            SVG path string
        """
        if context is None:
            buffer = context = Path()
        else:
            buffer = None

        points = self._clip(i)
        if points is None or len(points) == 0:
            return
        context.move_to(points[0], points[1])
        n = len(points)
        while points[0] == points[n - 2] and points[1] == points[n - 1] and n > 1:
            n -= 2
        for i in range(2, n, 2):
            if points[i] != points[i - 2] or points[i + 1] != points[i - 1]:
                context.line_to(points[i], points[i + 1])
        context.close_path()
        return None if buffer is None else str(buffer)

    def cell_polygons(self) -> Iterator[list[tuple[float, float]] | None]:
        """
        Returns an iterable over the non-empty polygons for each cell, with the
        cell index as property.

        Returns
        -------
        Iterator[list[tuple[float, float]] | None]
            Non-empty polygons for each cell
        """
        points = self.delaunay.points
        for i in range(len(points) // 2):
            cell = self.cell_polygon(i)
            if cell:
                # cell.index = i # lists cannot have attributes
                yield cell

    def cell_polygon(self, i: int) -> list[tuple[float, float]] | None:
        """
        Returns the convex closed polygon :math:`[[x_0, y_0], [x_1, y_1],
        \\ldots, [x_0, y_0]]` representing the cell for the specified point
        :code:`i`.

        Parameters
        ----------
        i : int
            Point index

        Returns
        -------
        list[tuple[float, float]] | None
            Convex closed polygon :math:`[[x_0, y_0], [x_1, y_1], \\ldots,
            [x_0, y_0]]`
        """
        polygon = Polygon()
        self.render_cell(i, polygon)
        return polygon.value()

    def _render_segment(
        self,
        x0: float,
        y0: float,
        x1: float,
        y1: float,
        context=None,
    ):
        c0 = self._regioncode(x0, y0)
        c1 = self._regioncode(x1, y1)
        if c0 == 0 and c1 == 0:
            context.move_to(x0, y0)
            context.line_to(x1, y1)
        elif S := self._clip_segment(x0, y0, x1, y1, c0, c1):
            context.move_to(S[0], S[1])
            context.line_to(S[2], S[3])

    def contains(self, i: int, x: float, y: float) -> bool:
        """
        Returns :code:`True` if the cell with the specified index :code:`i`
        contains the specified point :math:`(x, y)`; i.e., whether the point
        :code:`i` is the closest point in the diagram to the specified point.
        (This method is not affected by the associated Voronoi diagram's
        viewport bounds.)

        Parameters
        ----------
        i : int
            Cell index
        x : float
            X-coordinate point
        y : float
            Y-coordinate point

        Returns
        -------
        bool
            :code:`True` if cell contains the specified point
        """
        if isnan(x) or isnan(y):
            return False
        return self.delaunay._step(i, x, y) == i

    def neighbors(self, i: int) -> Iterator[int]:
        """
        Returns an iterable over the indexes of the cells that share a common
        edge with the specified cell :code:`i`. Voronoi neighbors are always
        neighbors on the Delaunay graph, but the converse is false when the
        common edge has been clipped out by the Voronoi diagram's viewport.

        Parameters
        ----------
        i : int
            Cell index

        Returns
        -------
        Iterator[int]
            Iterable over the indexes of the cells
        """
        ci = self._clip(i)
        if not ci:
            return
        for j in self.delaunay.neighbors(i):
            cj = self._clip(j)
            if not cj:
                continue

            break_loop = False
            li = len(ci)
            for ai in range(0, li, 2):
                lj = len(cj)
                for aj in range(0, lj, 2):
                    if (
                        ci[ai] == cj[aj]
                        and ci[ai + 1] == cj[aj + 1]
                        and ci[(ai + 2) % li] == cj[(aj + lj - 2) % lj]
                        and ci[(ai + 3) % li] == cj[(aj + lj - 1) % lj]
                    ):
                        yield j
                        break_loop = True
                        break
                if break_loop:
                    break

    def _cell(self, i: int) -> list[float] | None:
        circumcenters = self.circumcenters
        inedges = self.delaunay.inedges
        halfedges = self.delaunay.halfedges
        triangles = self.delaunay.triangles

        e0 = inedges[i]
        if e0 == -1:
            return None

        points = []
        e = e0
        while True:
            t = floor(e / 3)
            points.extend([circumcenters[t * 2], circumcenters[t * 2 + 1]])
            e = e - 2 if e % 3 == 2 else e + 1
            if triangles[e] != i:
                break
            e = halfedges[e]
            if e == e0 or e == -1:
                break
        return points

    def _clip(self, i: int) -> tuple[float, float, float, float, float, float]:
        if i == 0 and len(self.delaunay.hull) == 1:
            return [
                self.xmax,
                self.ymin,
                self.xmax,
                self.ymax,
                self.xmin,
                self.ymax,
                self.xmin,
                self.ymin,
            ]

        points = self._cell(i)
        if points is None:
            return None

        vectors = self.vectors
        v = i * 4
        return self._simplify(
            self._clip_infinite(
                i, points, vectors[v], vectors[v + 1], vectors[v + 2], vectors[v + 3]
            )
            if vectors[v] or vectors[v + 1]
            else self._clip_finite(i, points)
        )

    def _clip_finite(
        self,
        i: int,
        points: list[float],
    ) -> tuple[float, float, float, float, float, float]:
        n = len(points)
        P = None
        x0 = None
        y0 = None
        x1 = points[n - 2]
        y1 = points[n - 1]
        c0 = None
        c1 = self._regioncode(x1, y1)
        e0 = None
        e1 = 0
        for j in range(0, n, 2):
            x0 = x1
            y0 = y1
            x1 = points[j]
            y1 = points[j + 1]
            c0 = c1
            c1 = self._regioncode(x1, y1)
            if c0 == 0 and c1 == 0:
                e0 = e1
                e1 = 0
                if P:
                    P.extend([x1, y1])
                else:
                    P = [x1, y1]
            else:
                if c0 == 0:
                    S = self._clip_segment(x0, y0, x1, y1, c0, c1)
                    if S is None:
                        continue
                    sx0, sy0, sx1, sy1 = S
                else:
                    S = self._clip_segment(x1, y1, x0, y0, c1, c0)
                    if S is None:
                        continue
                    sx1, sy1, sx0, sy0 = S
                    e0 = e1
                    e1 = self._edgecode(sx0, sy0)
                    if e0 and e1:
                        self._edge(i, e0, e1, P, len(P))
                    if P:
                        P.extend([sx0, sy0])
                    else:
                        P = [sx0, sy0]
                e0 = e1
                e1 = self._edgecode(sx1, sy1)
                if e0 and e1:
                    self._edge(i, e0, e1, P, len(P))
                if P:
                    P.extend([sx1, sy1])
                else:
                    P = [sx1, sy1]
        if P:
            e0 = e1
            e1 = self._edgecode(P[0], P[1])
            if e0 and e1:
                self._edge(i, e0, e1, P, len(P))
        elif self.contains(
            i,
            (self.xmin + self.xmax) * 0.5,
            (self.ymin + self.ymax) * 0.5,
        ):
            return [
                self.xmax,
                self.ymin,
                self.xmax,
                self.ymax,
                self.xmin,
                self.ymax,
                self.xmin,
                self.ymin,
            ]
        return P

    def _clip_segment(
        self,
        x0: float,
        y0: float,
        x1: float,
        y1: float,
        c0: float,
        c1: float,
    ) -> tuple[float, float, float, float] | None:
        flip = c0 < c1
        if flip:
            x0, y0, x1, y1, c0, c1 = x1, y1, x0, y0, c1, c0

        while True:
            if c0 == 0 and c1 == 0:
                return [x1, y1, x0, y0] if flip else [x0, y0, x1, y1]
            if c0 & c1:
                return None
            c = c0 or c1
            if c & 0b1000:
                x = x0 + (x1 - x0) * (self.ymax - y0) / (y1 - y0)
                y = self.ymax
            elif c & 0b0100:
                x = x0 + (x1 - x0) * (self.ymin - y0) / (y1 - y0)
                y = self.ymin
            elif c & 0b0010:
                y = y0 + (y1 - y0) * (self.xmax - x0) / (x1 - x0)
                x = self.xmax
            else:
                y = y0 + (y1 - y0) * (self.xmin - x0) / (x1 - x0)
                x = self.xmin
            if c0:
                x0 = x
                y0 = y
                c0 = self._regioncode(x0, y0)
            else:
                x1 = x
                y1 = y
                c1 = self._regioncode(x1, y1)

    def _clip_infinite(
        self,
        i: int,
        points: list[float],
        vx0: float,
        vy0: float,
        vxn: float,
        vyn: float,
    ) -> tuple[float, float, float, float, float, float]:
        P = list(points)
        p = self._project(P[0], P[1], vx0, vy0)
        if p:
            P.insert(0, p[1])
            P.insert(0, p[0])
        p = self._project(P[-2], P[-1], vxn, vyn)
        if p:
            P.append(p[0])
            P.append(p[1])
        P = self._clip_finite(i, P)
        if P:
            c1 = self._edgecode(P[-2], P[-1])
            n = len(P)
            j = 0
            while j < n:
                c0 = c1
                c1 = self._edgecode(P[j], P[j + 1])
                if c0 and c1:
                    j = self._edge(i, c0, c1, P, j)
                    n = len(P)
                j += 2
        elif self.contains(
            i, (self.xmin + self.xmax) * 0.5, (self.ymin + self.ymax) * 0.5
        ):
            P = [
                self.xmin,
                self.ymin,
                self.xmax,
                self.ymin,
                self.xmax,
                self.ymax,
                self.xmin,
                self.ymax,
            ]
        return P

    def _edge(self, i: int, e0: float, e1: float, P: list[float], j: int) -> int:
        while e0 != e1:
            match e0:
                case 0b0101:
                    e0 = 0b0100
                    continue
                case 0b0100:
                    e0 = 0b0110
                    x = self.xmax
                    y = self.ymin
                case 0b0110:
                    e0 = 0b0010
                    continue
                case 0b0010:
                    e0 = 0b1010
                    x = self.xmax
                    y = self.ymax
                case 0b1010:
                    e0 = 0b1000
                    continue
                case 0b1000:
                    e0 = 0b1001
                    x = self.xmin
                    y = self.ymax
                case 0b1001:
                    e0 = 0b0001
                    continue
                case 0b0001:
                    e0 = 0b0101
                    x = self.xmin
                    y = self.ymin

            if (j >= len(P) or P[j] != x or P[j + 1] != y) and self.contains(i, x, y):
                P.insert(j, x)
                P.insert(j + 1, y)
                j += 2
        return j

    def _project(
        self, x0: float, y0: float, vx: float, vy: float
    ) -> tuple[float, float] | None:
        t = inf
        x = nan
        y = nan
        if vy < 0:
            if y0 <= self.ymin:
                return None
            c = (self.ymin - y0) / vy
            if c < t:
                y = self.ymin
                t = c
                x = x0 + t * vx
        elif vy > 0:
            if y0 >= self.ymax:
                return None
            c = (self.ymax - y0) / vy
            if c < t:
                y = self.ymax
                t = c
                x = x0 + t * vx

        if vx > 0:
            if x0 >= self.xmax:
                return None
            c = (self.xmax - x0) / vx
            if c < t:
                x = self.xmax
                t = c
                y = y0 + t * vy
        elif vx < 0:
            if x0 <= self.xmin:
                return None
            c = (self.xmin - x0) / vx
            if c < t:
                x = self.xmin
                t = c
                y = y0 + t * vy

        return [x, y]

    def _edgecode(self, x: float, y: float) -> int:
        if x == self.xmin:
            a = 0b0001
        elif x == self.xmax:
            a = 0b0010
        else:
            a = 0b0000

        if y == self.ymin:
            b = 0b0100
        elif y == self.ymax:
            b = 0b1000
        else:
            b = 0b0000

        return a | b

    def _regioncode(self, x: float, y: float) -> int:
        if x < self.xmin:
            a = 0b0001
        elif x > self.xmax:
            a = 0b0010
        else:
            a = 0b0000

        if y < self.ymin:
            b = 0b0100
        elif y > self.ymax:
            b = 0b1000
        else:
            b = 0b0000

        return a | b

    def _simplify(self, P: list[float]) -> list[float] | None:
        if P and len(P) > 4:
            n = len(P)
            i = 0
            while i < n:
                j = (i + 2) % n
                k = (i + 4) % n
                if (
                    P[i] == P[j]
                    and P[j] == P[k]
                    or P[i + 1] == P[j + 1]
                    and P[j + 1] == P[k + 1]
                ):
                    P.pop(j)
                    P.pop(j)
                    n = len(P)
                else:
                    i += 2
            if not n:
                P = None
        return P



from collections.abc import Callable

from ..color import hsl as color_hsl
from .color import color, hue


class HSLInterpolator:
    def __init__(self, func):
        self.func = func

    def __call__(self, start, end):
        start = color_hsl(start)
        end = color_hsl(end)
        h = self.func(start.h, end.h)
        s = color(start.s, end.s)
        l = color(start.l, end.l)
        opacity = color(start.opacity, end.opacity)

        def interpolate(t):
            start.h = h(t)
            start.s = s(t)
            start.l = l(t)
            start.opacity = opacity(t)
            return str(start)

        return interpolate


def interpolate_hsl(a: str, b: str) -> Callable[[float], str]:
    """
    Returns an HSL color space interpolator between the two colors a and b.
    The colors a and b need not be in HSL; they will be converted to HSL
    using d3.hsl. If either color's hue or saturation is NaN, the opposing
    color's channel value is used. The shortest path between hues is used.
    The return value of the interpolator is an RGB string.

    Parameters
    ----------
    a : str
        String a
    b : str
        String b

    Returns
    -------
    Callable[[float], str]
        Interpolator function

    Examples
    --------

    >>> interpolator = d3.interpolate_hsl("red", "blue")
    >>> interpolator(0)
    'rgb(255, 0, 0)'
    >>> interpolator(1)
    'rgb(0, 0, 255)'
    >>> interpolator(0.5)
    'rgb(255, 0, 255)'
    """
    return HSLInterpolator(hue)(a, b)


def interpolate_hsl_long(a: str, b: str) -> Callable[[float], str]:
    """
    Like :code:`interpolate_hsl`, but does not use the shortest path between hues.

    Parameters
    ----------
    a : str
        String a
    b : str
        String b

    Returns
    -------
    Callable[[float], str]
        Interpolator function

    Examples
    --------

    >>> interpolator = d3.interpolate_hsl_long("red", "blue")
    >>> interpolator(0)
    'rgb(255, 0, 0)'
    >>> interpolator(1)
    'rgb(0, 0, 255)'
    >>> interpolator(0.5)
    'rgb(0, 255, 0)'
    """
    return HSLInterpolator(color)(a, b)



import math

from ...selection import Selection
from .common import Curve, isvaluable


class StepCurve(Curve):
    def __init__(self, context, t):
        self._context = context
        self._t = t
        self._line = math.nan
        self._x = math.nan
        self._y = math.nan

    def area_start(self):
        self._line = 0

    def area_end(self):
        self._line = math.nan

    def line_start(self):
        self._x = math.nan
        self._y = math.nan
        self._point = 0

    def line_end(self):
        if 0 < self._t < 1 and self._point == 2:
            self._context.line_to(self._x, self._y)
        if isvaluable(self._line) or (self._line != 0 and self._point == 1):
            self._context.close_path()
        if self._line >= 0:
            self._t = 1 - self._t
            self._line = 1 - self._line

    def point(self, x, y):
        if self._point == 0:
            self._point = 1
            if isvaluable(self._line):
                self._context.line_to(x, y)
            else:
                self._context.move_to(x, y)
        elif self._point == 1:
            self._point = 2
            if self._t <= 0:
                self._context.line_to(self._x, y)
                self._context.line_to(x, y)
            else:
                x1 = self._x * (1 - self._t) + x * self._t
                self._context.line_to(x1, self._y)
                self._context.line_to(x1, y)
        else:
            if self._t <= 0:
                self._context.line_to(self._x, y)
                self._context.line_to(x, y)
            else:
                x1 = self._x * (1 - self._t) + x * self._t
                self._context.line_to(x1, self._y)
                self._context.line_to(x1, y)
        self._x = x
        self._y = y


def curve_step(context: Selection) -> Curve:
    """
    Produces a piecewise constant function (a step function) consisting of
    alternating horizontal and vertical lines. The y-value changes at the
    midpoint of each pair of adjacent x-values.

    Parameters
    ----------
    context : Selection
        Context

    Returns
    -------
    Curve
        Curve object

    Examples
    --------

    >>> points = [
    ...     [2, 2],
    ...     [6, 8],
    ...     [10, 10],
    ...     [12, 10],
    ...     [14, 4],
    ...     [20, 4],
    ...     [24, 8],
    ...     [29, 6],
    ...     [32, 4],
    ...     [35, 5],
    ...     [38, 2],
    ... ]
    >>> line = d3.line().curve(d3.curve_step)
    >>> line(points)
    'M2,2L4,2L4,8L8,8L8,10L11,10L11,10L13,10L13,4L17,4L17,4L22,4L22,8L26.500,8L26.500,6L30.500,6L30.500,4L33.500,4L33.500,5L36.500,5L36.500,2L38,2'

    **Result**

    .. image:: ../../figures/light_curve_step.svg
       :align: center
       :class: only-light

    .. image:: ../../figures/dark_curve_step.svg
       :align: center
       :class: only-dark
    """
    return StepCurve(context, 0.5)


def curve_step_before(context: Selection) -> Curve:
    """
    Produces a piecewise constant function (a step function) consisting of
    alternating horizontal and vertical lines. The y-value changes before the
    x-value.

    Parameters
    ----------
    context : Selection
        Context

    Returns
    -------
    Curve
        Curve object

    Examples
    --------

    >>> points = [
    ...     [2, 2],
    ...     [6, 8],
    ...     [10, 10],
    ...     [12, 10],
    ...     [14, 4],
    ...     [20, 4],
    ...     [24, 8],
    ...     [29, 6],
    ...     [32, 4],
    ...     [35, 5],
    ...     [38, 2],
    ... ]
    >>> line = d3.line().curve(d3.curve_step_before)
    >>> line(points)
    'M2,2L2,8L6,8L6,10L10,10L10,10L12,10L12,4L14,4L14,4L20,4L20,8L24,8L24,6L29,6L29,4L32,4L32,5L35,5L35,2L38,2'

    **Result**

    .. image:: ../../figures/light_curve_step_before.svg
       :align: center
       :class: only-light

    .. image:: ../../figures/dark_curve_step_before.svg
       :align: center
       :class: only-dark
    """
    return StepCurve(context, 0.0)


def curve_step_after(context: Selection) -> Curve:
    """
    Produces a piecewise constant function (a step function) consisting of
    alternating horizontal and vertical lines. The y-value changes after the
    x-value.

    Parameters
    ----------
    context : Selection
        Context

    Returns
    -------
    Curve
        Curve object

    Examples
    --------

    >>> points = [
    ...     [2, 2],
    ...     [6, 8],
    ...     [10, 10],
    ...     [12, 10],
    ...     [14, 4],
    ...     [20, 4],
    ...     [24, 8],
    ...     [29, 6],
    ...     [32, 4],
    ...     [35, 5],
    ...     [38, 2],
    ... ]
    >>> line = d3.line().curve(d3.curve_step_after)
    >>> line(points)
    'M2,2L6,2L6,8L10,8L10,10L12,10L12,10L14,10L14,4L20,4L20,4L24,4L24,8L29,8L29,6L32,6L32,4L35,4L35,5L38,5L38,2'

    **Result**

    .. image:: ../../figures/light_curve_step_after.svg
       :align: center
       :class: only-light

    .. image:: ../../figures/dark_curve_step_after.svg
       :align: center
       :class: only-dark
    """
    return StepCurve(context, 1.0)
